
# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'leftEQUIVALENCErightIMPLIESleftORleftANDrightNOTAND ATOM EQUIVALENCE FALSE IMPLIES LPAREN NOT OR RPAREN TRUEformula : ATOM\n                   | TRUE\n                   | FALSEformula : NOT formulaformula : formula OR formulaformula : formula AND formulaformula : formula IMPLIES formulaformula : formula EQUIVALENCE formulaformula : LPAREN formula RPAREN'
    
_lr_action_items = {'EQUIVALENCE':([1,3,4,6,11,12,13,14,15,16,17,],[8,-2,-1,-3,8,-4,-7,-8,-5,-6,-9,]),'IMPLIES':([1,3,4,6,11,12,13,14,15,16,17,],[7,-2,-1,-3,7,-4,7,7,-5,-6,-9,]),'RPAREN':([3,4,6,11,12,13,14,15,16,17,],[-2,-1,-3,17,-4,-7,-8,-5,-6,-9,]),'OR':([1,3,4,6,11,12,13,14,15,16,17,],[9,-2,-1,-3,9,-4,9,9,-5,-6,-9,]),'LPAREN':([0,2,5,7,8,9,10,],[2,2,2,2,2,2,2,]),'TRUE':([0,2,5,7,8,9,10,],[3,3,3,3,3,3,3,]),'AND':([1,3,4,6,11,12,13,14,15,16,17,],[10,-2,-1,-3,10,-4,10,10,10,-6,-9,]),'ATOM':([0,2,5,7,8,9,10,],[4,4,4,4,4,4,4,]),'NOT':([0,2,5,7,8,9,10,],[5,5,5,5,5,5,5,]),'$end':([1,3,4,6,12,13,14,15,16,17,],[0,-2,-1,-3,-4,-7,-8,-5,-6,-9,]),'FALSE':([0,2,5,7,8,9,10,],[6,6,6,6,6,6,6,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'formula':([0,2,5,7,8,9,10,],[1,11,12,13,14,15,16,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> formula","S'",1,None,None,None),
  ('formula -> ATOM','formula',1,'p_formula_atom','pl.py',61),
  ('formula -> TRUE','formula',1,'p_formula_atom','pl.py',62),
  ('formula -> FALSE','formula',1,'p_formula_atom','pl.py',63),
  ('formula -> NOT formula','formula',2,'p_formula_not','pl.py',72),
  ('formula -> formula OR formula','formula',3,'p_formula_or','pl.py',76),
  ('formula -> formula AND formula','formula',3,'p_formula_and','pl.py',80),
  ('formula -> formula IMPLIES formula','formula',3,'p_formula_implies','pl.py',84),
  ('formula -> formula EQUIVALENCE formula','formula',3,'p_formula_equivalence','pl.py',88),
  ('formula -> LPAREN formula RPAREN','formula',3,'p_formula_expression','pl.py',92),
]
